@@ -1,4 +1,20 @@
-/* Released to the public domain in 2003 by Sam Trenholme */
+/* Copyright (c) 2002 Sam Trenholme
+ *
+ * TERMS
+ *
+ * Redistribution and use in source and binary forms, with or without
+ * modification, are permitted provided that the following conditions
+ * are met:
+ *
+ * 1. Redistributions of source code must retain the above copyright
+ *    notice, this list of conditions and the following disclaimer.
+ * 2. Redistributions in binary form must reproduce the above copyright
+ *    notice, this list of conditions and the following disclaimer in the
+ *    documentation and/or other materials provided with the distribution.
+ *
+ * This software is provided 'as is' with no guarantees of correctness or
+ * fitness for purpose.
+ */
 
 #include "../libs/JsStr.h"
 #include "../MaraDns.h"
@@ -8,8 +24,14 @@
 /* Maximum allowed number of dlabel points */
 #define MAX_DLABEL_POINTS 512
 
-/* Maximum allowed length of compressed string */
+/* Maximum allowed length of compressed string; this is 4096 for TCP 
+ * packets */
+#ifdef AUTHONLY
+#define MAX_COMPRESSED_LEN 4096
+#else
 #define MAX_COMPRESSED_LEN 512
+#endif
+
 
 /* We use three "private" functions in Decompress.c */
 extern js_string 
@@ -61,7 +83,7 @@ compress_state *compress_init_state(js_string *uncompressed) {
     new->uncompressed = uncompressed;
 
     /* Initialize the values which allocate memory */
-    if((new->compressed = js_create(MAX_COMPRESSED_LEN,1)) == 0) {
+    if((new->compressed = js_create(MAX_COMPRESSED_LEN + 3,1)) == 0) {
         js_dealloc(new);
 	return 0;
 	}
