@@ -292,7 +292,7 @@ int decompress_dname(js_string *compressed, js_string *uncompressed,
    output: JS_ERROR on error, JS_SUCCESS on success
 */
 
-int compress_data(js_string *uncompressed, js_string *compressed) {
+int legacy_compress_data(js_string *uncompressed, js_string *compressed) {
     uint16 type,newrdlength,rdlength;
     int counter,place,cplace,ccount,rdptr,temp;
     q_header hdr;
