@@ -1,32 +1,31 @@
-/* Placed in the public domain by Sam Trenholme 2002 */
+/* Copyright (c) 2002 Sam Trenholme
+ *
+ * TERMS
+ *
+ * Redistribution and use in source and binary forms, with or without
+ * modification, are permitted provided that the following conditions
+ * are met:
+ *
+ * 1. Redistributions of source code must retain the above copyright
+ *    notice, this list of conditions and the following disclaimer.
+ * 2. Redistributions in binary form must reproduce the above copyright
+ *    notice, this list of conditions and the following disclaimer in the
+ *    documentation and/or other materials provided with the distribution.
+ *
+ * This software is provided 'as is' with no guarantees of correctness or
+ * fitness for purpose.
+ */
 
 #include <stdio.h>
 #include "../MaraDns.h"
-#include "Compress_rrs.h" /* The description of RRs to decompress */
+#include "../server/timestamp.h"
+#include "Compress_rrs.h" 
 #include "Compress_rrdescs.h" /* The description of RRs to decompress */
-
-/* Data structure to store the RRs in an easy-to-use format */
-/*typedef struct rrdesc {
-    int rr_num;
-    char *description;
-    struct rrdesc *next;
-    } rrdesc; */
+#include "functions_dns.h"
 
 /* Some definitions */
 #define MAX_DLABEL_LEN 256
 #define MAX_RR_SECTIONS 16
-#define RR_HASH_SIZE 7
-/* Various field types in the resource record */
-#define RRSUB_DLABEL 64
-#define RRSUB_TEXT 65
-#define RRSUB_VARIABLE 66
-
-/* Some defines to make -Wall happy */
-extern int show_timestamp();
-extern int js_readuint16();
-extern int legacy_decompress_data();
-extern int dlabel_length();
-extern int show_esc_stdout();
 
 /* The hash that stores RRs */
 rrdesc **rr_formats;
@@ -561,6 +560,12 @@ int decomp_add_rrdesc(js_string *desc) {
         new->description[counter] = 0;
 	}
 
+    /* Initialize the "tocompress" field to uninitialized; this is
+       a positive number under 127 because some versions of GCC on
+       some architectures consider 'char' without 'unsigned' an
+       unsigned value. */
+    new->tocompress = 79;
+
     /* Now, parse the string describing the message */
     fieldnum = subfieldnum = 1;
     rtype = -1;
@@ -649,10 +654,21 @@ int decomp_add_rrdesc(js_string *desc) {
 	     }    
          else if(fieldnum == 4) { /* Whether we can compress this field 
 	                             or not; currently ignored */
-	    if(c == '|') {
+             if(c == 'C' && new->tocompress == 79) {
+	        new->tocompress = 1;
+		}
+             else if(c == 'N' && new->tocompress == 79) {
+	        new->tocompress = 0;
+		}
+	     else if(c == '|' && new->tocompress != 79) {
                 subfieldnum = 1;
 	        fieldnum = 5;
 		}
+             else { /* Invalid for field num */
+	        js_dealloc(new->description);
+		js_dealloc(new);
+		return -4;
+		}
 	     }
          else if(fieldnum == 5) { /* Description of field; currently
 	                             ignored */
@@ -668,8 +684,8 @@ int decomp_add_rrdesc(js_string *desc) {
 
      place = rtype % RR_HASH_SIZE;
      if(rr_formats == 0) {
-	 js_dealloc(new->description);
-         js_dealloc(new);
+        js_dealloc(new->description);
+        js_dealloc(new);
         return JS_ERROR;
         }
      point = rr_formats[place];
@@ -933,9 +949,15 @@ int decomp_decompress_packet(js_string *compressed, js_string *uncompressed) {
         return JS_ERROR;
 
     /* Create the string for storing the rddata */
+#ifndef AUTHONLY
     if((rddata = js_create(512,1)) == 0) {
         return JS_ERROR;
 	}
+#else
+    if((rddata = js_create(4512,1)) == 0) {
+        return JS_ERROR;
+	}
+#endif
     /* Read the header */
 
     answers = decomp_get_header(compressed,uncompressed);
