@@ -1,26 +1,43 @@
-/* Placed in the public domain 2001 by Sam Trenholme */
+/* Copyright (c) 2002-2006 Sam Trenholme
+ *
+ * TERMS
+ *
+ * Redistribution and use in source and binary forms, with or without
+ * modification, are permitted provided that the following conditions
+ * are met:
+ *
+ * 1. Redistributions of source code must retain the above copyright
+ *    notice, this list of conditions and the following disclaimer.
+ * 2. Redistributions in binary form must reproduce the above copyright
+ *    notice, this list of conditions and the following disclaimer in the
+ *    documentation and/or other materials provided with the distribution.
+ *
+ * This software is provided 'as is' with no guarantees of correctness or
+ * fitness for purpose.
+ */
 
 /* Simple testing DNS resolving client.  This sends out simple
    queries to DNS servers and prints out to stdout the replies
    it receives.
 */
 
-#include "../libs/MaraHash.h"
 #include <stdio.h>
 #include <unistd.h>
 #include <stdlib.h>
 #include <errno.h>
 #include <string.h>
 #include <time.h>
+#ifndef MINGW32
 #include <sys/types.h>
 #include <sys/time.h>
 #include <sys/socket.h>
 #include <netinet/in.h>
 #include <arpa/inet.h>
+#else /* MINGW32 */
+#include <winsock.h>
+#include <wininet.h>
+#endif /* MINGW32 */
 #include "../MaraDns.h"
-#include "../dns/functions_dns.h"
-#include "../parse/functions_parse.h"
-
 /* All of the labels */
 #include "askmara_labels_en.h"
 /* BEGIN RNG USING CODE */
@@ -28,18 +45,22 @@
 #include "../rng/rng-api-fst.h"
 /* END RNG USING CODE */
 
+#include "../libs/MaraHash.h"
+#include "../dns/functions_dns.h"
+#include "../parse/functions_parse.h"
+
+int verbose_output(js_string *uindata);
+int csv2_compatible_output(js_string *uindata, js_string *query);
+int out_answer(js_string *uindata,int *place);
+int csv2_answer(js_string *uindata, int *place, js_string *query);
+/* escape_stdout_csv2: From parse/Csv2_esc_txt.c and defined in
+   parse/Csv2_functions.h */
+extern int escape_stdout_csv2(js_string *js);
+
 int verbose_mode = 0; /* Whether to have short or verbose output; 
                          1: verbose output; any other value: short output */
 int timeout = 21;
 
-/* So -Wall doesn't gag */
-extern int decomp_init();
-
-int csv1_answer(js_string *uindata, int *place, js_string *query);
-int csv1_compatible_output(js_string *uindata, js_string *query);
-int verbose_output(js_string *uindata);
-int out_answer(js_string *uindata,int *place);
-
 /* Octal escape any non-ASCII sequences in the askmara output; almost idential
    to show_esc_stdout */
 int escape_stdout(js_string *js) {
@@ -71,7 +92,7 @@ int escape_stdout(js_string *js) {
 u_int16_t gen_id(char *hostname) {
     /* BEGIN RNG USING CODE */
     /* May as well bring out the bug guns (e.g. our secure RNG) */
-    BYTE r_inBlock[17],r_outBlock[17],r_binKey[17];
+    MARA_BYTE r_inBlock[17],r_outBlock[17],r_binKey[17];
     keyInstance r_keyInst;
     cipherInstance r_cipherInst;
     unsigned char crypto_key[34];
@@ -98,7 +119,7 @@ u_int16_t gen_id(char *hostname) {
     /*memset(crypto_key,'ñ',16);*/
     /* Make the first 16 characters of the hostname the key (padded by the
        ñ character--do you see a n~ there?) */
-    strncpy(crypto_key,hostname,16);
+    strncpy((char *)crypto_key,hostname,16);
     /* Set the plaintext block to 'ññññññññññññññññ' (UTF-8 encoding) */
     mc2 = mc1;
     for(counter = 0; counter < 16; counter++) {
@@ -110,10 +131,10 @@ u_int16_t gen_id(char *hostname) {
     /*memset(r_inBlock,'ñ',16);*/
     /* If the hostname is longer than 16 characters, make the plaintext
        the longer part of the hostname */
-    if(js_strnlen(hostname,1000000) > 16)
-    strncpy(r_inBlock,hostname+16,16);
+    if(strlen(hostname) > 16)
+    strncpy((char *)r_inBlock,hostname+16,16);
 
-    if(makeKey(&r_keyInst, DIR_ENCRYPT, 128, crypto_key) != 1)
+    if(makeKey(&r_keyInst, DIR_ENCRYPT, 128, (char *)crypto_key) != 1)
         return JS_ERROR;
     if(cipherInit(&r_cipherInst, MODE_ECB, NULL) != 1)
         return JS_ERROR;
@@ -147,10 +168,19 @@ int main(int argc, char **argv) {
     struct timeval tv;  /* select() */
     int n; /* Select() return value */
     int nrd = 0; /* whether recursion is desired or not 1: No recursion;
-                   0: recursion desired */
+                    0: recursion desired */
     int id;
     int attempts = 10;
+    int desired_rr = -1; /* This is used to determine what RR type the
+			    user desires */
+    char *temp;
 
+#ifdef MINGW32
+    int err;
+    WSADATA wsaData;
+    WORD wVersionRequested = MAKEWORD(2,2);
+    err = WSAStartup(wVersionRequested, &wsaData);
+#endif /* MINGW32 */
     /* Determine what the query string is */
     if(argc < 2 || argc > 6)
         harderror(L_USAGE);
@@ -178,13 +208,13 @@ int main(int argc, char **argv) {
     if(argc == 3 + verbose_mode + nrd)
         server_address = argv[2 + verbose_mode + nrd];
     else
-        server_address = "127.0.0.3";
+        server_address = ASKMARA_DEFAULT_SERVER;
 
-    if((indata = js_create(512,1)) == 0)
+    if((indata = js_create(4512,1)) == 0)
        harderror(L_JS_CREATE_INDATA);
-    if((uindata = js_create(2048,1)) == 0)
+    if((uindata = js_create(12048,1)) == 0)
        harderror(L_JS_CREATE_UINDATA);
-    if((outdata = js_create(512,1)) == 0)
+    if((outdata = js_create(4512,1)) == 0)
        harderror(L_JS_CREATE_OUTDATA);
     if((question.qname = js_create(512,1)) == 0)
        harderror(L_JS_CREATE_QNAME);
@@ -225,7 +255,32 @@ int main(int argc, char **argv) {
     /* Create a DNS question , and put it in raw UDP format */
     /* DNS question -> looking up an A record... */
     question.qclass = 1; /* ...on the internet */
-    if(js_qstr2js(question.qname,argv[1 + verbose_mode + nrd]) == JS_ERROR)
+    /* We now allow people to specify any numeric record in this format:
+     * 40:t1.example.com., where 40 is the numeric rrtype */
+    temp = argv[1 + verbose_mode + nrd];
+    if(temp == 0) {
+	    printf("Wrong number of arguments to askmara\n");
+	    exit(43);
+    }
+    if(*temp >= '0' && *temp <= '9') {
+	int td = 0;
+	desired_rr = 0;
+	for(td = 0; td < 7; td++) {
+		if(*temp < '0' || *temp > '9') {
+			break;
+		}
+		desired_rr *= 10;
+		desired_rr += *temp - '0';
+		temp++;
+	}
+	if(*temp != ':') {
+		printf("Invalid numeric rtype\n");
+		exit(44);
+	}
+	*temp = 'A';
+    }
+
+    if(js_qstr2js(question.qname,temp) == JS_ERROR)
         harderror(L_INVALID_Q); /* Invalid query */
 
     /* Make sure that the csv1-compatible output hashes all non-RR
@@ -241,6 +296,10 @@ int main(int argc, char **argv) {
     if(place == JS_ERROR)
         harderror(L_INVALID_DQ); /* Invalid form of domain query */
     question.qtype = place;
+    /* 40:name.example.com. style queries */
+    if(desired_rr >= 0 && desired_rr <= 65535) {
+	    question.qtype = desired_rr;
+    }
 
     /* Make a string containing the RAW UDP query */
     make_hdr(&header,outdata);
@@ -264,7 +323,8 @@ int main(int argc, char **argv) {
         if(n == 0) /* Timeout */
             harderror("Timeout");
         if((count = recvfrom(s,indata->string,indata->max_count,0,
-                        (struct sockaddr *)&dns_udp,&len_inet)) < 0)
+                        (struct sockaddr *)&dns_udp,
+                        (socklen_t *)&len_inet)) < 0)
             harderror(L_DNS_R_ERROR); /* Problem getting DNS server response */
 
         indata->unit_count = count;
@@ -294,7 +354,7 @@ int main(int argc, char **argv) {
     if(verbose_mode == 1)
         verbose_output(uindata);
     else
-        csv1_compatible_output(uindata,question.qname);
+        csv2_compatible_output(uindata,question.qname);
 
     return 0;
     }
@@ -414,9 +474,9 @@ int out_answer(js_string *uindata,int *place) {
             printf("%s%s",L_NEWLINE,L_S_RNAME); /* SOA rname */
             escape_stdout(soa.rname);
             printf("%s%s%u%s",L_NEWLINE,L_S_SERIAL,soa.serial,L_NEWLINE);
-            printf("%s%d%s",L_S_REFRESH,soa.refresh,L_NEWLINE);
-            printf("%s%d%s",L_S_RETRY,soa.retry,L_NEWLINE);
-            printf("%s%d%s",L_S_EXPIRE,soa.expire,L_NEWLINE);
+            printf("%s%d%s",L_S_REFRESH,(int)soa.refresh,L_NEWLINE);
+            printf("%s%d%s",L_S_RETRY,(int)soa.retry,L_NEWLINE);
+            printf("%s%d%s",L_S_EXPIRE,(int)soa.expire,L_NEWLINE);
             printf("%s%u%s",L_S_MINIMUM,soa.minimum,L_NEWLINE);
             }
         else if(rr_hdr.type == RR_MX) {
@@ -459,16 +519,16 @@ int out_answer(js_string *uindata,int *place) {
 
         *place += rr_hdr.rdlength; /* To do: read the RD data itself for
                                       all MaraDNS supported RRs */
-        return 0;
-	}
+	return 0;
+        }
 
-/* csv1-compatible form of askmara output
+/* csv2-compatible form of askmara output
    input: Pointer to js_string object with uncompressed UDP data,
           Raw UDP for of the question
    output: None, shows data on standard output
 */
 
-int csv1_compatible_output(js_string *uindata, js_string *query) {
+int csv2_compatible_output(js_string *uindata, js_string *query) {
 
     int place,counter,count;
     q_question question;
@@ -500,17 +560,17 @@ int csv1_compatible_output(js_string *uindata, js_string *query) {
 
     /* Get all of the an replies */
     for(counter = 0; counter < header.ancount; counter++)
-        csv1_answer(uindata,&place,query);
+        csv2_answer(uindata,&place,query);
 
     /* Get all of the ns replies */
     printf("# %s%s",L_NSREP,L_NEWLINE);
     for(counter = 0; counter < header.nscount; counter++)
-        csv1_answer(uindata,&place,query);
+        csv2_answer(uindata,&place,query);
 
     /* Get all of the ar replies */
     printf("# %s%s",L_ARREP,L_NEWLINE);
     for(counter = 0; counter < header.arcount; counter++)
-        csv1_answer(uindata,&place,query);
+        csv2_answer(uindata,&place,query);
 
     return 1;
     }
@@ -523,13 +583,15 @@ int csv1_compatible_output(js_string *uindata, js_string *query) {
    output: outputs to stdout the data in that string
 */
 
-int csv1_answer(js_string *uindata, int *place, js_string *query) {
+int csv2_answer(js_string *uindata, int *place, js_string *query) {
 
     q_rr rr_hdr;
     rr_soa soa;
     rr_mx mx;
     int count;
     js_string *lower1, *lower2;
+    js_string *t;
+    int rtype;
 
     /* Create the needed strings */
     if((rr_hdr.name = js_create(512,1)) == 0)
@@ -562,32 +624,32 @@ int csv1_answer(js_string *uindata, int *place, js_string *query) {
     if(js_issame(lower1,lower2) != 1)
         printf("#");
 
+    rtype = rr_hdr.type;
     hname_translate(rr_hdr.name,rr_hdr.type);
-    escape_stdout(rr_hdr.name);
-    /* We also show the type if this is an unsupported record type */
-    if(rr_hdr.type != RR_A &&
-       rr_hdr.type != RR_NS &&
-       rr_hdr.type != RR_CNAME &&
-       rr_hdr.type != RR_SOA &&
-       rr_hdr.type != RR_PTR &&
-       rr_hdr.type != RR_MX &&
-       rr_hdr.type != RR_TXT)
-        printf("|%d",rr_hdr.type);
-    printf("|%u",rr_hdr.ttl);
+    t = js_create(256,1);
+    if(js_copy(rr_hdr.name,t) == JS_ERROR) {
+	    harderror("Couldn't copy hostname");
+    }
+    t->string++;
+    t->unit_count--;
+    escape_stdout(t);
+    t->string--;
+    t->unit_count++;
+    js_destroy(t);
     if(rr_hdr.type == RR_SOA) {
         if(read_soa(uindata,&soa,*place) == JS_ERROR)
             harderror(L_READSOA);
-        hname_translate(soa.mname,RR_SOA);
-        hname_translate(soa.rname,RR_MX);
-        printf("|"); /* SOA mname */
+        hname_translate(soa.mname,RR_MAGIC_SPACE);
+        hname_translate(soa.rname,RR_MAGIC_EMAIL);
+        printf(" +%u",rr_hdr.ttl);
+        printf(" soa"); /* SOA mname */
         escape_stdout(soa.mname);
-        printf("|"); /* SOA rname */
         escape_stdout(soa.rname);
-        printf("|%u",soa.serial);
-        printf("|%d",soa.refresh);
-        printf("|%d",soa.retry);
-        printf("|%d",soa.expire);
-        printf("|%u\n",soa.minimum);
+        printf(" %u",soa.serial);
+        printf(" %d",(int)soa.refresh);
+        printf(" %d",(int)soa.retry);
+        printf(" %d",(int)soa.expire);
+        printf(" %u\n",soa.minimum);
         }
     else if(rr_hdr.type == RR_MX) {
         if(uindata->unit_count < *place + 2)
@@ -595,15 +657,28 @@ int csv1_answer(js_string *uindata, int *place, js_string *query) {
         mx.preference = ((*(uindata->string + *place) & 0xff) << 8) |
                          (*(uindata->string + *place + 1) & 0xff);
         read_ns(uindata,mx.exchange,*place + 2);
-        printf("|%d",mx.preference); /* MX pref. */
-        hname_translate(mx.exchange,-2); /* Makes the char a pipe */
+        printf(" +%u mx %d",rr_hdr.ttl,mx.preference); /* MX pref. */
+        hname_translate(mx.exchange,RR_MAGIC_SPACE); /* Makes the char a 
+                                                      * space */
         escape_stdout(mx.exchange);
         printf("\n");
         }
     else if(rr_hdr.type == RR_NS || rr_hdr.type == RR_CNAME ||
             rr_hdr.type == RR_PTR) {
+        printf(" +%u",rr_hdr.ttl);
+        switch(rr_hdr.type) {
+		case RR_NS:
+			printf(" ns");
+			break;
+		case RR_CNAME:
+			printf(" cname");
+			break;
+		case RR_PTR:
+			printf(" ptr");
+			break;
+		}
         read_ns(uindata,mx.exchange,*place);
-        hname_translate(mx.exchange,-2); /* First char is a pipe */
+        hname_translate(mx.exchange,RR_MAGIC_SPACE); /* First char is space */
         escape_stdout(mx.exchange);
         printf("\n");
         }
@@ -611,23 +686,50 @@ int csv1_answer(js_string *uindata, int *place, js_string *query) {
         if(uindata->unit_count < *place + 4)
             return JS_ERROR;
         /* Display the IP of the data */
-        printf("|%d%s%d%s%d%s%d\n",*(uindata->string + *place),
+        printf(" +%u a %d%s%d%s%d%s%d\n", rr_hdr.ttl,
+			*(uindata->string + *place),
                L_DOT,*(uindata->string + *place + 1),
                L_DOT,*(uindata->string + *place + 2),
                L_DOT,*(uindata->string + *place + 3));
         }
     else if(rr_hdr.type == RR_TXT) {
-        printf("|"); 
-        read_txt(uindata,mx.exchange,*place);
-        escape_stdout(mx.exchange);
-        printf("\n");
-        }
+        int len,p;
+	len = rr_hdr.rdlength;
+	p = *place;
+        printf(" +%u txt ",rr_hdr.ttl); 
+	/* XXX: This doesn't correctly handle multi-chunked TXT data */
+	while(len > 0) {
+            int q;
+            q = read_txt(uindata,mx.exchange,p);
+            q++;
+            if(q < 0) {
+                printf("Problem reading TXT record\n");
+                exit(1);
+                }
+            escape_stdout_csv2(mx.exchange);
+	    len -= q;
+            p += q;
+	    if(len > 0) { 
+                printf(";");
+                }
+            else {
+                printf("\n");
+                }
+            }
+         }
     else {
-        printf("|\n");
+	printf(" +%u raw %d ",rr_hdr.ttl,rr_hdr.type);
+        if(js_substr(uindata,mx.exchange,*place,rr_hdr.rdlength) == JS_ERROR) {
+		printf("Problem copying string over\n");
+		exit(55);
+	}
+        escape_stdout_csv2(mx.exchange);
+        printf(" \n");
 	}
 
     *place += rr_hdr.rdlength; /* To do: read the RD data itself for
                                   all MaraDNS supported RRs */
     return 0;
     }
+    
 
